
import os
import json
import shlex
from jinja2 import Template
from typing import List, Dict
from argparse import ArgumentParser
from rkd.contract import TaskInterface, ExecutionContext
from rkd.standardlib.shell import BaseShellCommandWithArgumentParsingTask
from collections import namedtuple


EnvironmentVariable = namedtuple('EnvironmentVariable', 'name value comment')


class DockerTagExistsTask(BaseShellCommandWithArgumentParsingTask):
    """ Check if a docker tag exists """

    def __init__(self):
        super().__init__(
            name=':tag-exists',
            group=':docker',
            description='',
            arguments_definition=None,
            command='''
                image_full_name="$ARG_IMAGE";
                wait_time="${ARG_WAIT_TIME:-10}"
                search_term='Pulling|is up to date|not found'
                
                result="$((timeout --preserve-status "$wait_time" docker 2>&1 pull "$image_full_name" &) | grep -v 'Pulling repository' | egrep -o "$search_term")"
                test "$result" || { echo "Timed out too soon. Try using a wait_time greater than $wait_time..."; return 1 ;}
                
                if echo $result | grep -vq 'not found'; then
                    echo "Image found."
                    exit 0
                else
                    echo "Image not found"
                    exit 1
                fi
            '''
        )

    def configure_argparse(self, parser: ArgumentParser):
        parser.add_argument('--image', '-i', required=True, help='Image name')
        parser.add_argument('--wait-time', '-w', default='10', help='Time to wait after which to interrupt the pulling')


class ExtractEnvsFromDockerfileTask(TaskInterface):
    """ Extract environment variables from a Dockerfile """

    def get_name(self) -> str:
        return ':extract-envs-from-dockerfile'

    def get_group_name(self) -> str:
        return ':docker'

    def execute(self, context: ExecutionContext) -> bool:
        out_format = context.args['format']
        file_path = context.args['file']

        self._io.out(self.extract(out_format, file_path))
        return True

    def extract(self, out_format: str, file_path: str) -> str:
        if not os.path.isfile(file_path):
            raise Exception('Cannot find Dockerfile at path "%s"' % file_path)

        with open(file_path, 'rb') as f:
            content = f.read().decode('utf-8')

        out_vars = self.get_envs(content)

        if out_format == 'bash_source':
            return 'export DOCKERFILE_ENVS=%s' % shlex.quote(json.dumps(out_vars))
        elif out_format == 'json':
            return json.dumps(out_vars, indent=2, sort_keys=True)
        else:
            buf = ''

            for name, out_var in out_vars.items():
                buf += "export %s=%s\n" % (out_var.name, out_var.value)

            return buf

    def configure_argparse(self, parser: ArgumentParser):
        parser.add_argument('--file', '-f', required=True, help='Path to Dockerfile to read as input')
        parser.add_argument('--format', default='json', help='Output format, one of: json, bash_source, env')

    def get_envs(self, content: str):
        by_lines = content.split("\n")
        normalized_lines = list(map(lambda curr: curr.strip(), by_lines))
        pos = -1
        matched_lines = {}

        for line in normalized_lines:
            pos += 1

            # each block that starts with "ENV" can be multi-line, so it will be extracted multi-line
            if line.upper().startswith('ENV'):
                matched_lines = {**matched_lines, **self._parse_lines(self._extract_block(pos, normalized_lines))}

        return matched_lines

    def _parse_lines(self, lines: List[str]) -> Dict[str, EnvironmentVariable]:
        parsed = {}
        comment_buffer = ''

        for line in lines:
            # if that's a comment or a comment continuation, then collect into the buffer
            if line.startswith('#'):
                comment_buffer += line.lstrip('# ') + "\n"
            else:
                # if the comment block ended, or it is not a comment block
                if line.startswith('ENV'):
                    line = line[3:].strip()

                sep = line.split('=', maxsplit=1)

                # potential syntax error
                if len(sep) < 2:
                    self._io.warn('!!! Cannot parse block in line "%s"' % line)
                    continue

                parsed[sep[0].strip()] = EnvironmentVariable(
                    name=sep[0].strip(),
                    value=sep[1].strip(),
                    comment=comment_buffer.strip()
                )
                comment_buffer = ''

        return parsed

    @staticmethod
    def _extract_block(start_pos: int, lines: List[str]) -> list:
        extracted = []
        pos = start_pos
        length = len(lines)

        extracted.append(lines[start_pos].rstrip('\\'))

        while True:
            pos += 1

            if pos > length:
                break

            line = lines[pos]

            if line.endswith('\\') or line.startswith('#'):
                extracted.append(line.rstrip('\\'))
                continue
            else:
                extracted.append(line)
                break

        return extracted


class GenerateReadmeTask(TaskInterface):
    """ Generate README.md.j2 into README.md considering env variables from Dockerfile """

    def get_name(self) -> str:
        return ':generate-readme'

    def get_group_name(self) -> str:
        return ':docker'

    def execute(self, context: ExecutionContext) -> bool:
        readme_path = context.args['target_path']
        readme_template_path = context.args['template']
        dockerfile_path = context.args['dockerfile']

        if not os.path.isfile(readme_template_path):
            self._io.error_msg('Path to template is not valid')
            return False

        if not os.path.isfile(dockerfile_path):
            self._io.error_msg('Path to Dockerfile is not valid')
            return False

        variables = self.extract_envs_from_dockerfile(dockerfile_path)

        with open(readme_template_path, 'rb') as f:
            rendered = Template(f.read().decode('utf-8')).render({'DOCKERFILE_ENVS': variables})

            if readme_path:
                with open(readme_path, 'w') as rw:
                    rw.write(rendered)
            else:
                self._io.out(rendered)

        return True

    def extract_envs_from_dockerfile(self, dockerfile_path: str):
        """ Extracts environment variables list, values and descriptions from the Dockerfile """

        task = ExtractEnvsFromDockerfileTask()
        self.copy_internal_dependencies(task)
        return json.loads(task.extract('json', dockerfile_path))

    def configure_argparse(self, parser: ArgumentParser):
        parser.add_argument('--template', '-rt', required=True,
                            help='Readme template path (in Jinja2 format)')
        parser.add_argument('--target-path', '-t', required=False, default='',
                            help='Path where to write the README. If not specified, then stdout will be preferred')
        parser.add_argument('--dockerfile', '-f', required=True, help='Path to the Dockerfile to parse')



import re
from subprocess import check_output, CalledProcessError, STDOUT
from typing import Union


class GitTools:
    @classmethod
    def get_current_tag(cls) -> Union[str, None]:
        try:
            return check_output(
                'git describe --exact-match --tags $(git log -n1 --pretty=\'%h\')',
                shell=True,
                stderr=STDOUT
            )\
                .decode('utf-8').strip()
        except CalledProcessError:
            return None

    @classmethod
    def is_currently_on_tag(cls) -> bool:
        return cls.get_current_tag() is not None

    @classmethod
    def get_last_tag(cls) -> str:
        return check_output(
                'git for-each-ref refs/tags --sort=-taggerdate --format=\'%(refname)\' --count=1',
                shell=True,
                stderr=STDOUT
        )\
            .decode('utf-8').strip().replace('refs/tags/', '')


class VersionTools:
    @classmethod
    def get_next_version(cls) -> str:
        return cls.increment_version(GitTools.get_last_tag())

    @classmethod
    def increment_version(cls, version: str) -> str:
        if not version:
            return '0.0.1'

        matches = re.match('([0-9.]+)', version)
        parts = ['0', '0', '1']

        # find first numeric part that contains a dot (to exclude eg. RC-1 or dev8, dev.12)
        for match in reversed(matches.groups()):
            if "." in match:
                parts = match.split('.')
                break

        # increment 3rd part (patch)
        if len(parts) >= 3:
            parts[2] = str(int(parts[2]) + 1)
        else:
            parts.append('1')

        return '.'.join(parts)



import re
from typing import Dict
from argparse import ArgumentParser
from subprocess import CalledProcessError
from rkd.contract import TaskInterface, ExecutionContext
from rkd.syntax import TaskDeclaration
from rkd.standardlib.docker import TagImageTask
from rkd.standardlib.docker import PushTask
from .github import ForEachGithubReleaseTask
from .github import FindClosestReleaseTask
from .docker import DockerTagExistsTask
from .tools import VersionTools, GitTools


class ProcessRequestTask(TaskInterface):
    """Takes incoming request from the CI

    Responsibility:
        - Decide about the destination version number and format
        - Decide if we REBUILD existing tags or not (release enforces this)
    """

    def get_name(self) -> str:
        return ':process'

    def get_group_name(self) -> str:
        return ':boat-ci'

    def execute(self, context: ExecutionContext) -> bool:
        self.io().h1('BOAT CI - Shifts everything into the Harbor')

        # input
        commit_message: str = context.args['commit_message']
        cmd_to_exec: str = context.args['exec']
        task_type: str = context.args['type']
        git_tag: str = GitTools.get_current_tag() if GitTools.is_currently_on_tag() else ''

        # logic
        rebuild = False
        tag_template = context.args['dev_version_template']

        if GitTools.is_currently_on_tag():
            self.io().h4('Rebuilding as currently on a tag')
            rebuild = True
            tag_template = context.args['release_version_template']

        if "@force-rebuild" in commit_message:
            self.io().h4('Rebuilding as @force-rebuild specified')
            rebuild = True

        if "@force-rebuild-last-tag" in commit_message:
            self.io().h4('Rebuilding last tag - @force-rebuild-last-tag specified')
            git_tag = GitTools.get_last_tag()
            tag_template = context.args['release_version_template']

        # Parsing
        tag_template = self._parse_template_str(
            tag_template,
            task_type=task_type,
            should_rebuild=rebuild,
            version='',
            git_tag=git_tag
        )

        self.io().h4('Boat-CI produced tag template: %s' % tag_template)

        cmd_to_exec = self._parse_template_str(
            cmd_to_exec,
            task_type=task_type,
            should_rebuild=rebuild,
            version=tag_template,
            git_tag=git_tag
        )

        try:
            self.sh('set -x; %s' % cmd_to_exec, verbose=True)
        except CalledProcessError as e:
            self.io().error_msg(str(e))
            return False

        return True

    def _parse_template_str(self, template: str, task_type: str, should_rebuild: bool, git_tag: str,
                            version: str = '') -> str:
        parsed = template

        if "%NEXT_VERSION%" in template:
            parsed = parsed.replace('%NEXT_VERSION%', VersionTools.get_next_version())

        parsed = parsed.replace('%REBUILD_FLAG%', '' if should_rebuild else ' --dont-rebuild-when-exists')

        if version:
            parsed = parsed.replace('%VERSION%', version)

        if task_type in ['each-release', 'specific-release']:
            task_type = self.get_group_name() + ':' + task_type

        parsed = parsed.replace('%RELEASE_TASK%', task_type)\
            .replace('%VERSION_TEMPLATE%', version)\
            .replace('%GIT_TAG%', git_tag)

        return parsed

    def configure_argparse(self, parser: ArgumentParser):
        # CI input
        parser.add_argument('--commit-message', required=True, help='Commit message')

        # Build mode
        parser.add_argument(
            '--type', '-t',
            default='each-release',
            help='Task type, or task name eg. each-release, specific-release. Pasted as %RELEASE_TASK% to the --exec'
        )
        parser.add_argument(
            '--exec', '-e',
            help='Build task name. Defaults to building all releases from github',
            default='rkd --no-ui %RELEASE_TASK% -rl info %REBUILD_FLAG% --version-template="%VERSION_TEMPLATE%"'
        )
        parser.add_argument('--dev-version-template', default='%MATCH_0%-SNAPSHOT',
                            help='Examples: %MATCH_0%-D%NEXT_VERSION%-SNAPSHOT')
        parser.add_argument('--release-version-template', default='%MATCH_0%-D%GIT_TAG%')


class EachRelease(TaskInterface):
    def get_name(self) -> str:
        return ':each-release'

    def get_group_name(self) -> str:
        return ':boat-ci'

    def execute(self, context: ExecutionContext) -> bool:
        # input
        allowed_tags_regexp: str = context.getenv('ALLOWED_TAGS_REGEXP')
        dest_docker_repo: str = context.getenv('DEST_DOCKER_REPO')
        max_versions: int = int(context.getenv('MAX_VERSIONS'))
        github_repository: str = context.getenv('GITHUB_REPOSITORY')
        version_template: str = context.args['version_template']
        rebuild: bool = not context.args['dont_rebuild_when_exists']
        version_build_cmd: str = context.getenv('VERSION_BUILD_CMD')\
            .replace('%VERSION_TEMPLATE%', version_template)\
            .replace('%IMAGE%', dest_docker_repo)

        opts = ''

        if not rebuild:
            opts += ' --dont-rebuild '

        cmd = ('set -x; rkd :github:for-each-release --repository=%s --exec="%s" ' +
               ' --dest-docker-repo="%s" --allowed-tags-regexp="%s" ' +
               '--release-tag-template="%s" --max-versions=%i %s') % (
            github_repository,
            version_build_cmd,
            dest_docker_repo,
            allowed_tags_regexp,
            version_template,
            max_versions,
            opts
        )

        try:
            self.sh(cmd)
        except CalledProcessError as e:
            self.io().error_msg(str(e))
            return False

        return True

    def get_declared_envs(self) -> Dict[str, str]:
        return {
            'ALLOWED_TAGS_REGEXP': 'v([0-9.]+)',
            'DEST_DOCKER_REPO': '',
            'MAX_VERSIONS': '2',
            'GITHUB_REPOSITORY': '',
            'VERSION_BUILD_CMD': 'rkd :boat-ci:specific-release --dockerfile=./Dockerfile ' +
                                 ' --dir=. --docker-version="%VERSION_TEMPLATE%" --app-version="%MATCH_0%" ' +
                                 ' --image=%IMAGE%'
        }

    def configure_argparse(self, parser: ArgumentParser):
        parser.add_argument(
            '--dont-rebuild-when-exists',
            action='store_true'
        )
        parser.add_argument('--version-template', required=True)


class SpecificRelease(TaskInterface):
    """ Performs a docker build & tag & push """

    def get_name(self) -> str:
        return ':specific-release'

    def get_group_name(self) -> str:
        return ':boat-ci'

    def get_declared_envs(self) -> Dict[str, str]:
        return {
            'DOCKER_BUILD_OPTS': '',
            'IMAGE': '',
            'DOCKERFILE': '',
            'DIR': ''
        }

    def execute(self, context: ExecutionContext) -> bool:
        # facts
        work_dir = context.get_arg_or_env('--dir')
        dockerfile_path = context.get_arg_or_env('--dockerfile')
        image = context.get_arg_or_env('--image')
        image_version = context.args['docker_version']
        app_version = context.args['app_version']
        opts = self._parse_opts(context.get_arg_or_env('--docker-build-opts'), app_version)

        # complete docker image address with version
        tag = image + ':' + image_version

        # build & tag & publish
        if not self.silent_sh(('docker build %s -f %s -t %s %s ' +
                               '--build-arg RKT_APP_VERSION="%s" --build-arg RKT_IMG_VERSION=%s') %
                              (work_dir, dockerfile_path, tag, opts, app_version, image_version),
                              verbose=True):
            self.io().error('Cannot build docker image')
            return False

        self.rkd([':docker:tag', '--image=%s' % tag, '--propagate', '-rl=debug'], verbose=True)
        self.rkd([':docker:push', '--image=%s' % tag, '--propagate', '-rl=debug'], verbose=True)

        return True

    def _parse_opts(self, template: str, app_version: str) -> str:
        """Templating - allows to append additional information to build args of the Docker image

        Example case:
            --build-arg FRONTEND_VERSION=%FIND_CLOSEST_RELEASE(taigaio/taiga-front-dist)%
            To find and paste a closest matching release of Taiga frontend, when building the backend.
            The algorithm is in :github:find-closest-release task and is always reproducible.
        """

        if '%FIND_CLOSEST_RELEASE' in template:
            all_matches = re.findall('%FIND_CLOSEST_RELEASE\((.*)\)%', template)

            for match in all_matches:
                template = template.replace(
                    '%%FIND_CLOSEST_RELEASE(%s)%%' % match,
                    self.sh(' '.join([
                        'rkd', '--silent',
                        ':github:find-closest-release',
                        '--repository=%s' % match,
                        '--compare-with=%s' % app_version
                    ]), capture=True).strip()
                )

        return template

    def configure_argparse(self, parser: ArgumentParser):
        parser.add_argument('--dockerfile', '-f', help='Path to Dockerfile')
        parser.add_argument('--dir', '-d', help='Build directory / build root')
        parser.add_argument('--docker-version', '-v', required=True, help='Version of the docker image (image tag)')
        parser.add_argument('--image', '-i', required=True, help='Image eg. quay.io/riotkit/tunman')
        parser.add_argument('--app-version', '-av', help='Application version')
        parser.add_argument('--docker-build-opts', '-o', default='',
                            help='Docker build opts eg. --build-arg SOME=THING')


def imports():
    return [
        TaskDeclaration(EachRelease()),
        TaskDeclaration(ProcessRequestTask()),
        TaskDeclaration(SpecificRelease()),

        # dependencies
        TaskDeclaration(ForEachGithubReleaseTask()),
        TaskDeclaration(FindClosestReleaseTask()),
        TaskDeclaration(TagImageTask()),
        TaskDeclaration(PushTask()),
        TaskDeclaration(DockerTagExistsTask())
    ]



from abc import ABC
from typing import Union, Pattern, Match

import requests
import time
import re
import subprocess
from argparse import ArgumentParser
from rkd.contract import TaskInterface, ExecutionContext


def natural_sort(l):
    convert = lambda text: int(text) if text.isdigit() else text.lower()
    alphanum_key = lambda key: [convert(c) for c in re.split('([0-9]+)', key)]

    return sorted(l, key=alphanum_key)


class BaseGithubTask(TaskInterface, ABC):
    def get_group_name(self) -> str:
        return ':github'

    def get_available_tags(self, url: str, sleep_time: int, retries: int = 5) -> list:
        """ Lists all tags from github project """

        tags_url = url + '/tags'
        self.io().h2('Getting latest github releases from %s' % tags_url)

        try:
            response = requests.get(tags_url).json()

            if "message" in response and response['message'] == 'Not Found':
                raise Exception('Repository on github not found')

            return list(map(
                lambda tag_object: str(tag_object['name']),
                response
            ))
        except TypeError:
            if retries <= 0:
                raise

            time.sleep(sleep_time)
            return self.get_available_tags(url, sleep_time, retries - 1)

    def configure_argparse(self, parser: ArgumentParser):
        parser.add_argument('--retries', '-r', default='5', help='Maximum number of retries in request to github')
        parser.add_argument('--retry-wait', '-w', default='5', help='Amount of seconds between retries')
        parser.add_argument('--repository', '-n', required=True, help='Repository name eg. riotkit-org/filerepository')


class FindClosestReleaseTask(BaseGithubTask):
    """ Find a github release that is closest to the selected number """

    def get_name(self) -> str:
        return ':find-closest-release'

    def execute(self, context: ExecutionContext) -> bool:
        url = 'https://api.github.com/repos/%s' % context.args['repository']

        self._io.out(self.find_closest_version(
            version=context.args['compare_with'],
            url=url,
            sleep_time=int(context.args['retry_wait']),
            retries=int(context.args['retries'])
        ))

        return True

    def configure_argparse(self, parser: ArgumentParser):
        super().configure_argparse(parser)
        parser.add_argument('--compare-with', '-c', required=True, help='Version to compare with')

    def find_closest_version(self, version: str, url: str, sleep_time: int, retries: int) -> str:
        all_versions = self.get_available_tags(url, sleep_time, retries)

        if version in all_versions:
            return version

        all_versions += [version]
        sorted_desc = list(reversed(natural_sort(all_versions)))
        current_version_position = sorted_desc.index(version)

        if len(sorted_desc) == 1:
            return version

        if current_version_position == len(sorted_desc) - 1:
            return sorted_desc[len(sorted_desc) - 2]

        return sorted_desc[current_version_position + 1]


class ForEachGithubReleaseTask(BaseGithubTask):
    """ Iterate over recent X github releases and execute a task """

    def get_name(self) -> str:
        return ':for-each-release'

    def execute(self, context: ExecutionContext) -> bool:
        self.io().h1('Iterating over each github release')

        url = 'https://api.github.com/repos/%s' % context.args['repository']
        force_rebuild = not context.args['dont_rebuild']
        tags = self.get_available_tags(
            url,
            int(context.args['retry_wait']),
            int(context.args['retries'])
        )
        self.io().h4('Release tag template is "%s"' % context.args['release_tag_template'])

        return self.print_last_versions(
            tags=tags,
            max_versions=int(context.args['max_versions']),
            allowed_tags_regexp=re.compile(context.args['allowed_tags_regexp']) \
            if context.args['allowed_tags_regexp'] else None,
            release_tag_template=context.args['release_tag_template'],
            force_rebuild=force_rebuild,
            build_command=context.args['exec'],
            dest_docker_repo=context.args['repository'],
            dry_run=bool(context.args['dry_run'])
        )

    def print_last_versions(self, tags: list, max_versions: int, allowed_tags_regexp: Union[Pattern, None],
                            release_tag_template: str, force_rebuild: bool, build_command: str,
                            dest_docker_repo: str, dry_run: bool):
        to_build = {}
        result = True

        self.io().print_opt_line()
        self.io().h1('Parsing GIT tags to find out those that are release tags')

        # collect items to build
        for tag in tags:
            properties = None

            if allowed_tags_regexp:
                matches = allowed_tags_regexp.match(tag)
                if not matches:
                    self.io().h2("Not matched tag \"%s\"" % tag)
                    continue

                self.io().h2('Matched %s' % tag)
                properties = matches

            to_build[tag] = properties

        processed = 0
        self.io().h1('Processing tags (max amount: %i)' % max_versions)

        for git_tag, matches in to_build.items():
            if 0 < max_versions <= processed:
                break

            release_tag = self.create_release_tag(git_tag, matches, release_tag_template)

            if not force_rebuild and self.was_already_built(dest_docker_repo, release_tag):
                self.io().h2('Skipping "%s" as the docker tag already exists' % release_tag)
                processed += 1
                continue

            command = self.render_template(build_command, git_tag, matches)

            self.io().h2('Calling generated command %s' % command)

            if not dry_run:
                try:
                    self.sh(command)
                except subprocess.CalledProcessError:
                    result = False

            processed += 1
            self.io().print_separator()
            self.io().print_opt_line()

        return result

    def create_release_tag(self, git_tag: str, matches: Union[Match, None], release_tag_template: str):
        return self.render_template(release_tag_template, git_tag, matches, False)

    def render_template(self, original_text: str, git_tag: str,
                        matches: [Union, Match, None], with_release_tag: bool = True):

        """ Inject variable values into the template ex. input: release-%GIT_TAG% output: release-1.0.5 """

        text = original_text \
            .replace('%GIT_TAG%', git_tag)

        groups = matches.groups() if matches else None

        if groups:
            for match_num in range(0, len(groups)):
                text = text.replace('%MATCH_' + str(match_num) + '%', groups[match_num])

        if with_release_tag:
            text = text.replace('%RELEASE_TAG%', self.create_release_tag(git_tag, matches, original_text))

        self.io().h3('render_template(): "%s" into "%s"' % (original_text, text))

        return text

    def was_already_built(self, image_name: str, docker_tag: str) -> bool:
        """ Checks if the docker tag was already pushed """

        self.io().h3('Checking if docker tag "%s" was already pushed' % docker_tag)

        try:
            self.rkd([':docker:tag-exists', '--name=%s:%s' % (image_name, docker_tag)])
            return True

        except subprocess.CalledProcessError as e:
            if e.returncode == 127:
                raise Exception(':docker-tag-exists command is not registered')

            return False

    def configure_argparse(self, parser: ArgumentParser):
        super().configure_argparse(parser)

        parser.description = "For each released version of application on github perform a command. " + \
                             "Example: rkd :github:for-each-release --repository=riotkit-org/file-repository " + \
                             "--exec 'echo \"[%MATCH_0%]\"' --dest-docker-repo quay.io/riotkit/file-repository " + \
                             "-tr 'v([0-9.]+)'"

        parser.add_argument('--dest-docker-repo',
                            help='Docker repository ex. quay.io/riotkit/phpbb',
                            required=True)
        parser.add_argument('--exec', '-e',
                            help='Command to execute. Variables: %%GIT_TAG%% (original git tag name), ' +
                                 '%%MATCH_0%% (regexp match-0), ' +
                                 '%%MATCH_N%% (regexp match-N), %%RELEASE_TAG%% ' +
                                 '(defined by --release-tag-template)',
                            default='echo "%%GIT_TAG%%"')
        parser.add_argument('--dont-rebuild', '-dr',
                            help='Do not build the same version twice ' +
                                 '(checks existence of a docker tag for --docker-repo)',
                            action='store_true')
        parser.add_argument('--allowed-tags-regexp', '-tr',
                            help='Optional regexp to filter tags (eg. release-([0-9.]+) or v([0-9.]+))')
        parser.add_argument('--release-tag-template', '-t',
                            help='Tag that will be pushed to docker registry. The same variables apply there as for ' +
                                 '--exec, except %%RELEASE_TAG%%',
                            default='%GIT_TAG%')
        parser.add_argument('--max-versions', '-mv',
                            help='Max versions to check/build',
                            default='5')
        parser.add_argument('--dry-run',
                            help='Print instead of performing',
                            action='store_true')
        parser.add_argument('--verbose',
                            help='Print all messages like --debug, but also perform',
                            action='store_true')
